'use strict';

const
  should = require('should'),
  proxyquire = require('proxyquire'),
  sinon = require('sinon');

describe('plugin implementation', function () {
  let
    Plugin,
    plugin,
    setPort,
    setBackend,
    onSpy = sinon.spy(),
    sendSpy = sinon.spy(),
    badId = 'aBadId',
    goodId = 'aGoodId',
    goodChannel = 'aGoodChannel',
    badChannel = 'aBadChannel';

  before(function () {
    // stubbing websocket
    Plugin = proxyquire('../lib/index', {
      'ws': {
        Server: function(config) {
          setPort = config.port;

          return {
            on: onSpy
          };
        }
      }
    });
  });

  beforeEach(function () {
    setPort = -1;
    setBackend = null;
    plugin = new Plugin();
    onSpy.reset();
    sendSpy.reset();
  });

  describe('#general', function () {
    it('should expose an init function', function () {
      should(plugin.init).be.a.Function();
    });
  });

  describe('#init', function () {
    let
      config = {port: 1234},
      context = {foo: 'bar'};

    it('should set internal properties correctly', function () {
      let
        ret = plugin.init(config, context, true);

      should(ret).be.eql(plugin);
      should(plugin.config).be.eql(config);
      should(plugin.context).be.eql(context);
      should(setPort).be.eql(1234);
      should(setBackend).be.null();
    });

    it('should setup a websocket server if not in dummy mode', function () {
      let ret = plugin.init(config, context, false);

      should(ret).be.eql(plugin);
      should(plugin.config).be.eql(config);
      should(plugin.context).be.eql(context);
      should(setPort).be.eql(1234);
      should(onSpy.firstCall.args[0]).be.eql('connection');
      should(onSpy.firstCall.args[1]).be.Function();
      should(onSpy.secondCall.args[0]).be.eql('error');
      should(onSpy.secondCall.args[1]).be.Function();
    });
  });

  describe('#onConnection', function () {
    let
      config = {port: 1234},
      onClientSpy = sinon.stub(),
      clientSocketMock = {
        on: onClientSpy,
        close: sinon.stub(),
        upgradeReq: {
          connection: {
            remoteAddress: 'ip'
          },
          headers: {
            'X-Foo': 'bar',
            'x-forwarded-for': '1.1.1.1,2.2.2.2'
          }
        }
      },
      newConnectionSpy = sinon.stub(),
      context = {
        accessors: {
          router: {
            newConnection: newConnectionSpy
          }
        },
        constructors: {
          ClientConnection: sinon.spy(function () { this.id = 'id'; }) // eslint-disable-line no-invalid-this
        },
        log: {
          error: sinon.spy()
        }
      };

    beforeEach(() => {
      onClientSpy.reset();
      clientSocketMock.close.reset();
      newConnectionSpy.reset();
    });

    it('should bind proper listeners', () => {
      let
        clientDisconnectionStub = sinon.stub(plugin, 'onClientDisconnection'),
        clientMessageStub = sinon.stub(plugin, 'onClientMessage');

      newConnectionSpy.returns(Promise.resolve({a: 'connection', id: 'id'}));
      plugin.init(config, context, false);

      plugin.onConnection(clientSocketMock);

      should(onClientSpy.callCount).be.eql(3);
      should(onClientSpy.firstCall.args[0]).be.eql('close');
      should(onClientSpy.firstCall.args[1]).be.Function();
      should(onClientSpy.secondCall.args[0]).be.eql('error');
      should(onClientSpy.secondCall.args[1]).be.Function();
      should(onClientSpy.thirdCall.args[0]).be.eql('message');
      should(onClientSpy.thirdCall.args[1]).be.Function();

      should(clientDisconnectionStub.callCount).be.eql(0);
      should(clientMessageStub.callCount).be.eql(0);
      onClientSpy.firstCall.args[1]();
      should(clientDisconnectionStub.callCount).be.eql(1);
      should(clientMessageStub.callCount).be.eql(0);
      onClientSpy.secondCall.args[1]();
      should(clientDisconnectionStub.callCount).be.eql(2);
      should(clientMessageStub.callCount).be.eql(0);
      onClientSpy.thirdCall.args[1]();
      should(clientDisconnectionStub.callCount).be.eql(2);
      should(clientMessageStub.callCount).be.eql(1);

      clientDisconnectionStub.reset();
      clientMessageStub.reset();
      should(Object.keys(plugin.connectionPool).length).be.eql(1);

      should(context.constructors.ClientConnection)
        .be.calledOnce()
        .be.calledWith('websocket', ['1.1.1.1', '2.2.2.2', 'ip'], {'x-forwarded-for': '1.1.1.1,2.2.2.2', 'X-Foo': 'bar'});

      should(plugin.connectionPool.id)
        .match({
          alive: true,
          socket: clientSocketMock,
          channels: []
        });
    });

    it('should reject and close the socket if creating a connection fails', () => {
      const
        error = new Error('test');

      newConnectionSpy.throws(error);

      plugin.init(config, context, false);

      plugin.onConnection(clientSocketMock);

      should(context.log.error)
        .be.calledWith('[plugin-websocket] Unable to register connection to the proxy\n%s', error.stack);

      should(onClientSpy.callCount).be.eql(0);
      should(clientSocketMock.close.called).be.true();
      should(clientSocketMock.close.calledWith(4503, 'foobar'));

    });
  });

  describe('#broadcast', function () {
    let
      config = {port: 1234},
      context = {foo: 'bar'};

    it('should do nothing if channel does not exist', function () {
      plugin.init(config, context, false);
      plugin.broadcast({
        channels: [badChannel],
        payload: {}
      });
      should(sendSpy.callCount).be.eql(0);
    });


    it('should call send if all conditions are met', function () {
      plugin.init(config, context, false);
      plugin.channels = {
        [goodChannel]: {
          [goodId]: true
        }
      };
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          socket: {
            send: sendSpy
          }
        }
      };

      plugin.broadcast({
        channels: [goodChannel],
        payload: {some: 'data'}
      });
      should(sendSpy.callCount).be.eql(1);
      should(sendSpy.firstCall.args[0]).be.eql(JSON.stringify({some: 'data', room: goodChannel}));
    });
  });

  describe('#notify', function () {
    let
      config = {port: 1234},
      context = {foo: 'bar'};

    it('should do nothing if id does not exist', function () {
      plugin.init(config, context, false);
      plugin.notify({
        id: badId,
        payload: {},
        channels: [goodChannel]
      });
      should(sendSpy.callCount).be.eql(0);
    });

    it('should call send if all conditions are met', function () {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          socket: {
            send: sendSpy
          }
        }
      };

      plugin.notify({
        connectionId: goodId,
        channels: [goodChannel],
        payload: {some: 'data'}
      });
      should(sendSpy.callCount).be.eql(1);
      should(sendSpy.firstCall.args[0]).be.eql(JSON.stringify({some: 'data', room: goodChannel}));
    });
  });

  describe('#joinChannel', function () {
    let
      config = {port: 1234},
      context = {foo: 'bar'};

    it('should do nothing if id does not exist', function () {
      plugin.init(config, context, false);
      plugin.joinChannel({
        connectionId: badId
      });
      should(plugin.channels).be.deepEqual({});
    });

    it('should add clientId to the channel if conditions are met', function () {
      plugin.init(config, context, false);
      plugin.channels = {
        [goodChannel]: { count: 0 }
      };
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          socket: {
            send: sendSpy
          },
          channels: []
        }
      };

      plugin.joinChannel({
        connectionId: goodId,
        channel: goodChannel
      });

      should(plugin.channels).be.deepEqual({
        [goodChannel]: { [goodId]: true, count: 1 }
      });
    });

    it('should create the channel entry add clientId to the channel if conditions are met and channel did not exist before', function () {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          socket: {
            send: sendSpy
          },
          channels: []
        }
      };
      plugin.joinChannel({
        connectionId: goodId,
        channel: goodChannel
      });
      should(plugin.channels).be.deepEqual({
        [goodChannel]: { [goodId]: true, count: 1 }
      });
    });
  });

  describe('#leaveChannel', function () {
    let
      config = {port: 1234},
      context = {foo: 'bar'};

    it('should do nothing if id does not exist', function () {
      plugin.init(config, context, false);
      plugin.leaveChannel({
        id: badId
      });
      should(plugin.channels).be.deepEqual({});
    });

    it('should do nothing if channel does not exist', function () {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]:  {
          alive: true
        }
      };
      plugin.leaveChannel({
        connectionId: goodId
      });
      should(plugin.channels).be.deepEqual({});
    });

    it('should do nothing if id is not in channel', function () {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          socket: {
            send: sendSpy
          },
          channels: []
        }
      };
      plugin.channels = {
        [goodChannel]: {[badId]: true }
      };
      plugin.leaveChannel({
        connectionId: goodId,
        channel: goodChannel
      });
      should(plugin.channels).be.deepEqual({
        [goodChannel]: {[badId]: true}
      });
    });

    it('should remove id from channel if conditions are met', function () {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          socket: {
            send: sendSpy
          },
          channels: [goodChannel]
        }
      };
      plugin.channels = {
        [goodChannel]: {[goodId]: true, [badId]: true, count: 2}
      };
      plugin.leaveChannel({
        connectionId: goodId,
        channel: goodChannel
      });

      should(plugin.channels).be.deepEqual({
        [goodChannel]: {[badId]: true, count: 1}
      });
      should(plugin.connectionPool[goodId].channels.length).be.eql(0);
    });

    it('should remove id from channel if conditions are met and remove channel if it is empty', function () {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          socket: {
            send: sendSpy
          },
          channels: []
        }
      };
      plugin.channels = {
        [goodChannel]: {[goodId]: true}
      };
      plugin.leaveChannel({
        connectionId: goodId,
        channel: goodChannel
      });
      should(plugin.channels).be.deepEqual({});
    });
  });

  describe('#onMessage', function () {
    let
      config = {port: 1234},
      fakeRequest = {aRequest: 'Object'},
      requestStub = sinon.stub().returns(fakeRequest),
      executeStub = sinon.stub().yields({requestId: 'foo', content: {requestId: 'foo'}}),
      context = {
        constructors: {Request: requestStub},
        accessors: {router: {execute: executeStub}},
        errors: {BadRequestError: sinon.stub().returns({badrequest: 'error'})}
      };

    beforeEach(() => {
      requestStub.reset();
      executeStub.reset();
    });

    it('should do nothing if the data is undefined', function () {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]: {
          alive: true
        }
      };
      plugin.onClientMessage(badId, undefined);
      should(executeStub.callCount).be.eql(0);
      should(requestStub.callCount).be.eql(0);
    });

    it('should do nothing if the client is unknown', function () {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]: {
          alive: true
        }
      };
      plugin.onClientMessage(badId, JSON.stringify('aPayload'));
      should(executeStub.callCount).be.eql(0);
      should(requestStub.callCount).be.eql(0);
    });

    it('should execute the request if client and packet are ok', function () {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          connection: 'aConnection',
          socket: {
            send: sendSpy
          },
          channels: []
        }
      };

      plugin.onClientMessage(goodId, JSON.stringify('aPayload'));

      should(requestStub)
        .be.calledOnce()
        .be.calledWith('aPayload', {
          connectionId: 'aGoodId',
          protocol: 'websocket'
        });

      should(executeStub)
        .be.calledOnce()
        .be.calledWith(fakeRequest);

      should(sendSpy)
        .be.calledOnce()
        .be.calledWith(JSON.stringify({
          requestId: 'foo',
          room: 'foo'
        }));
    });

    it('should forward an exception to the client if the incoming message is not in JSON format', () => {
      plugin.init(config, context, false);
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          connection: 'aConnection',
          socket: {
            send: sendSpy
          },
          channels: []
        }
      };

      plugin.onClientMessage(goodId, 'foobar');
      should(requestStub.called).be.false();
      should(executeStub.called).be.false();
      should(sendSpy.calledOnce).be.true();
      should(sendSpy.firstCall.args[0]).be.eql(JSON.stringify({badrequest: 'error'}));
    });
  });

  describe('#onServerError', function () {
    let
      config = {port: 1234},
      context = {
        log: {
          error: sinon.spy()
        }
      };

    it('should log the error', () => {
      const error = new Error('test');

      plugin.init(config, context, false);
      plugin.onServerError(error);

      should(context.log.error)
        .be.calledOnce()
        .be.calledWith('[plugin-websocket] An error has occured:\n' + error.stack);
    });
  });

  describe('#onClientDisconnection', function () {
    let
      config = {port: 1234},
      removeConnectionSpy = sinon.stub().returns(Promise.resolve({a: 'connection'})),
      context = {accessors: {router: {removeConnection: removeConnectionSpy}}};

    beforeEach(() => {
      removeConnectionSpy.reset();
    });

    it('should do nothing if the client is unknown', function () {
      plugin.init(config, context, false);
      plugin.channels = {
        [goodChannel]: []
      };
      plugin.connectionPool = {
        [goodId]: {
          alive: true
        }
      };
      plugin.socketPool = {
        [goodId]: true
      };
      plugin.onClientDisconnection(badId);
      should(removeConnectionSpy.callCount).be.eql(0);
      should(plugin.connectionPool[goodId].alive).be.true();
      should(plugin.socketPool[goodId]).be.true();
    });

    it('should remove the client connection if it exists', function () {
      plugin.init(config, context, false);
      plugin.channels = {
        [goodChannel]: {[goodId]: true, 'foobar': true, count: 2}
      };
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          connection: 'aConnection',
          socket: {
            send: sendSpy
          },
          channels: [goodChannel]
        }
      };

      plugin.onClientDisconnection(goodId);
      should(removeConnectionSpy.callCount).be.eql(1);
      should(plugin.connectionPool).be.deepEqual({});
      should(plugin.channels).be.deepEqual({[goodChannel]: {foobar: true, count: 1}});
    });

    it('should remove a channel entirely if the last connection leaves', function () {
      plugin.init(config, context, false);
      plugin.channels = {
        [goodChannel]: {[goodId]: true}
      };
      plugin.connectionPool = {
        [goodId]: {
          alive: true,
          connection: 'aConnection',
          socket: {
            send: sendSpy
          },
          channels: [goodChannel]
        }
      };

      plugin.onClientDisconnection(goodId);
      should(removeConnectionSpy.callCount).be.eql(1);
      should(plugin.connectionPool).be.deepEqual({});
      should(plugin.channels).be.deepEqual({});
    });
  });

  describe('#disconnect', () => {
    it('should close the client connection', () => {
      plugin.connectionPool.id = {
        socket: {
          close: sinon.spy()
        }
      };

      plugin.disconnect('id');

      should(plugin.connectionPool.id.socket.close)
        .be.calledOnce()
        .be.calledWith('CLOSEDONREQUEST', 'Connection closed by remote host');
    });

  });
});
